﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace LayoutTransformControlSample
{
    /// <summary>
    /// Class to create a test matrix for LayoutTransformControl testing
    /// </summary>
    public static class LayoutTransformControlTesting
    {
        // Create a test matrix for LayoutTransformControl
        public static IEnumerable<LayoutTransformControl> CreateTestMatrix(Grid root, Transform transform)
        {
            var layoutTransformControls = new List<LayoutTransformControl>();
            var column = 0;
            var row = 0;
            // For each transformation type (RenderTransform, LayoutTransform, LayoutTransformControl)
            foreach (var transformer in new Func<FrameworkElement, FrameworkElement>[] {
                element => { element.RenderTransform = transform; return element; },
#if !SILVERLIGHT
                element => { element.LayoutTransform = transform; return element; },
#endif
                element =>
                {
                    var layoutTransformControl = new LayoutTransformControl { Child = element };
                    layoutTransformControl.Transform = transform;
                    layoutTransformControls.Add(layoutTransformControl);
                    return layoutTransformControl;
                },
            })
            {
                column = 1;
                // For each test element type (TextBlock, ContentControl, Button)
                foreach (var constructor in new Func<FrameworkElement>[] {
                    () => {
                        return new TextBlock {
                            Text = "TextBlock",
#if !SILVERLIGHT
                            Background = Brushes.LightGray,
#endif
                          }; },
                    () => {
                        var text = new TextBlock {
                            Text = "Content",
                            HorizontalAlignment = HorizontalAlignment.Center,
                            VerticalAlignment = VerticalAlignment.Center,
                        };
                        var border = new Border {
                            BorderBrush = new SolidColorBrush(Colors.Red),
                            BorderThickness = new Thickness(1),
                            Child = text,
                            Background = new SolidColorBrush(Colors.LightGray),
                        };
                        return new ContentControl {
                            Content = border,
                            HorizontalContentAlignment = HorizontalAlignment.Stretch,
                            VerticalContentAlignment = VerticalAlignment.Stretch,
                            IsTabStop = false,
                        };
                    },
                    // () => { return new Button { Content = "Button" }; },
                })
                {
                    // For each scenario (various constraint types)
                    foreach (var scenario in new Func<FrameworkElement, FrameworkElement>[] {
                        element => element,
                        element => Center(element, true, true),
                        element => Center(element, true, false),
                        element => Center(element, false, true),
                        element => WrapInTightGrid(element, true, true, 0, 0),
                        element => WrapInTightGrid(element, true, false, 0, 0),
                        element => WrapInTightGrid(element, false, true, 0, 0),
                        element => WrapInTightGrid(element, true, true, 0, 20),
                        element => WrapInTightGrid(element, true, true, 20, 0),
                    })
                    {
                        // Get a suitable element and add it to the test matrix
                        var frameworkElement = scenario(transformer(constructor()));
                        Grid.SetColumn(frameworkElement, column);
                        Grid.SetRow(frameworkElement, row);
                        root.Children.Add(frameworkElement);
                        column++;
                    }
                }
                row++;
            }
            // Add column/row definitions for the test elements (and and extra empty one at the end for grid lines)
            for (var i = 1; i < column; i++)
            {
                root.ColumnDefinitions.Add(new ColumnDefinition());
            }
            root.ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
            for (var i = 0; i < row; i++)
            {
                root.RowDefinitions.Add(new RowDefinition());
            }
            root.RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto });
            // Return collection of LayoutTransformControls
            return layoutTransformControls;
        }

        // Helper method to center the specified element
        private static FrameworkElement Center(FrameworkElement element, bool horizontally, bool vertically)
        {
            if (horizontally)
            {
                element.HorizontalAlignment = HorizontalAlignment.Center;
            }
            if (vertically)
            {
                element.VerticalAlignment = VerticalAlignment.Center;
            }
            return element;
        }

        // Helper method to wrap the specified element in a "tight" Grid with grid lines
        private static Grid WrapInTightGrid(FrameworkElement element, bool horizontally, bool vertically, double fixedWidth, double fixedHeight)
        {
            var grid = new Grid { ShowGridLines = true };
            grid.ColumnDefinitions.Add(new ColumnDefinition());
            if (horizontally)
            {
                grid.ColumnDefinitions.Add(new ColumnDefinition { Width = ((0 == fixedWidth) ? GridLength.Auto : new GridLength(fixedWidth)) });
                grid.ColumnDefinitions.Add(new ColumnDefinition());
            }
            grid.RowDefinitions.Add(new RowDefinition());
            if (vertically)
            {
                grid.RowDefinitions.Add(new RowDefinition { Height = ((0 == fixedHeight) ? GridLength.Auto : new GridLength(fixedHeight)) });
                grid.RowDefinitions.Add(new RowDefinition());
            }
            Grid.SetColumn(element, 1);
            Grid.SetRow(element, 1);
            grid.Children.Add(element);
            return grid;
        }
    }
}
